/*
 *  UFTP - UDP based FTP with multicast
 *
 *  Copyright (C) 2001-2010   Dennis A. Bush, Jr.   bush@tcnj.edu
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 3 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License along
 *  with this program; if not, write to the Free Software Foundation, Inc.,
 *  51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
 */

#include "uftp.h"

#define DEF_RATE 1000
#define DEF_WEIGHT 300
#define DEF_MIN_TIME 10
#define DEF_PORT 1044
#define DEF_PUB_MULTI "230.4.4.1"
#define DEF_PRIV_MULTI "230.5.5.x"
#define MAX_INTERFACES 20
#define USAGE "uftp [ -U ] [ -R txrate ] [ -W txweight ] [ -m min_time ]\n\
    [ -n ] [ -z ] [ -l latency_level ] [ -c client_wait ] [ -L logfile ]\n\
    [ -t ttl ] [ -I interface ] [ -p port ] [ -H host[,host...] ]\n\
    [ -B buf_size ] [ -M pub_multicast_addr ] [ -P priv_multicast_addr ] file\n"

struct destinfo_t {
  int done;
  struct in_addr addr;
  char name[50];
  timestamp time;
  int naks;
} destlist[MAXOPENDEST];

SOCKET sock;
int file,weight,min_time,destcount,unicast,verbose,latency_lev,noname,minimal_resp;
char filename[MAXFILENAME],*naklist;
uint32_t txID;
int32_t numblocks,numsections,rate,packet_wait,client_wait;
f_offset_t filesize;
struct sockaddr_in listen_dest,receive_dest;
FILE *applog;

const int latency_lev_count=4;
int latency[][6]={
/* time in millilseconds to wait for each state (first set are defaults) */
/* first four are for transmitter, last two for receiver */
/* register, late register, status, late completion, reg_conf, done_conf */
{2500,3500,4000,2000,3000,3000},
{500,1000,500,500,800,800},
{1000,2000,1500,1500,1400,1750},
{2500,3500,4000,4000,3000,5000},
{5000,7000,8000,8000,6000,10000}
};

extern char *optarg;
extern int optind;

void cleanup(void)
{
  closesocket(sock);
  if (file>=0) close(file);
  if (naklist!=NULL)
    free(naklist);

#ifdef WINDOWS
  WSACleanup();
#endif
}

void abort_msg(char errstr[], struct sockaddr_in *dest)
{
  char buf[PACKETSIZE],*data;
  struct uftp_h *header;

  data=&(buf[sizeof(struct uftp_h)]);
  header=(struct uftp_h *)&(buf[0]);
  header->uftp_id=UFTP_ID;
  header->func=ABORT;
  header->tx_id=txID;
  strncpy(data,errstr,BLOCKSIZE-1);
  data[BLOCKSIZE-1]='\x0';
  head_hton(header);
  if (sendto(sock,buf,sizeof(buf),0,(struct sockaddr *)dest,sizeof(struct sockaddr_in))==SOCKET_ERROR) {
    logfunc(applog,0,1,errno,1,"Error sending ABORT");
  }
}

void send_regconf(int attempt)
{
  int i,j,confcnt;
  char buf[PACKETSIZE];
  struct uftp_h *header;
  struct statusinfo *info;

  header=(struct uftp_h *)&(buf[0]);
  info=(struct statusinfo *)&(buf[sizeof(struct uftp_h)]);
  bzero(buf,sizeof(buf));
  header->uftp_id=UFTP_ID;
  header->func=REG_CONF;
  header->tx_id=txID;
  head_hton(header);
  confcnt=0;
  for (i=0,j=0;i<destcount;i++) {
    if (destlist[i].done==0) {
      info->addr_list[j++]=destlist[i].addr;
    } 
    if ((j>=MAXINFODEST)||((i==destcount-1)&&(j>0))) {
      logfunc(applog,0,0,0,0,"Sending REG_CONF %d%c\n",attempt-1,confcnt+'a');
      if (sendto(sock,buf,sizeof(buf),0,(struct sockaddr *)&receive_dest,sizeof(listen_dest))==SOCKET_ERROR) {
        logfunc(applog,0,1,errno,1,"Error sending REG_CONF");
        break;
      } 
      if (packet_wait) usleep(packet_wait);
      bzero(info->addr_list,MAXINFODEST*sizeof(struct in_addr));
      j=0;
      confcnt++;
    }
  }
}

void send_doneconf(int attempt)
{
  int i,j,confcnt;
  char buf[PACKETSIZE];
  struct uftp_h *header;
  struct statusinfo *info;

  header=(struct uftp_h *)&(buf[0]);
  info=(struct statusinfo *)&(buf[sizeof(struct uftp_h)]);
  bzero(buf,sizeof(buf));
  header->uftp_id=UFTP_ID;
  header->func=DONE_CONF;
  header->tx_id=txID;
  head_hton(header);
  confcnt=0;
  for (i=0,j=0;i<destcount;i++) {
    if (destlist[i].done==1) {
      info->addr_list[j++]=destlist[i].addr;
    } 
    if ((j>=MAXINFODEST)||((i==destcount-1)&&(j>0))) {
      logfunc(applog,0,0,0,0,"Sending DONE_CONF %d%c\n",attempt-1,confcnt+'a');
      if (sendto(sock,buf,sizeof(buf),0,(struct sockaddr *)&receive_dest,sizeof(listen_dest))==SOCKET_ERROR) {
        logfunc(applog,0,1,errno,1,"Error sending DONE_CONF");
        break;
      } 
      if (packet_wait) usleep(packet_wait);
      bzero(info->addr_list,MAXINFODEST*sizeof(struct in_addr));
      j=0;
      confcnt++;
    }
  }
}

void announce()
{
  struct sockaddr_in sin;
  fd_set fdin;
  struct timeval tv, tv_sav;
  struct uftp_h *header;
  struct fileinfo *info;
  struct hostent *hp;
  char buf[PACKETSIZE],*data;
  int rval,attempt,i,j,isclosed,resend,gotall,found,polltime,confcnt;
  socklen_t addr_len;

  header=(struct uftp_h *)&(buf[0]);
  info=(struct fileinfo *)&(buf[sizeof(struct uftp_h)]);
  data=&(buf[sizeof(struct uftp_h)]);
  isclosed=(destcount!=0);
  if (isclosed) {
    for (i=0;i<destcount;i++)
      destlist[i].done=-3;
  }
  polltime=latency[latency_lev][0];
  tv.tv_sec=polltime/1000;
  tv.tv_usec=(polltime%1000)*1000;
  tv_sav=tv;
  attempt=1;
  resend=1;
  gotall=0;
  while (attempt<=5) {
    if (resend) {
      send_regconf(attempt);
      bzero(buf,sizeof(buf));
      header->uftp_id=UFTP_ID;
      header->func=ANNOUNCE;
      header->tx_id=txID;
      info->block_total=numblocks;
      info->section_total=numsections;
      if (filesize>(f_offset_t)0x7FFFFFFF) {
        info->fsize=0;
        info->hilargefsize=(int32_t)(filesize >> 32);
        info->lolargefsize=(int32_t)(filesize & 0xFFFFFFFF);
      } else {
        info->fsize=(int32_t)filesize;
        info->hilargefsize=0;
        info->lolargefsize=0;
      }
      info->mcast.s_addr=unicast?0:receive_dest.sin_addr.s_addr;
      info->rxlatency1=(client_wait>latency[latency_lev][4])?client_wait:latency[latency_lev][4];
      info->rxlatency2=(client_wait>latency[latency_lev][5])?client_wait:latency[latency_lev][5];
      info->open=(isclosed==0);
      strncpy(info->name,basename(filename),sizeof(info->name)-1);
      info->minimal_resp=minimal_resp;
      info->name[sizeof(info->name)-1]='\x0';
      head_hton(header);
      info_hton(info);
      if (isclosed) {
        /* build client list for closed group and send one or more ANNOUNCE */
        confcnt=0;
        for (i=0,j=0;(isclosed)&&(i<destcount);i++) {
          if (destlist[i].done==-3) {
            info->addr_list[j++]=destlist[i].addr;
          }
          if ((j>=MAXDEST)||((i==destcount-1)&&(j>0))) {
            logfunc(applog,0,0,0,0,"Sending ANNOUNCE %d%c\n",attempt,confcnt+'a');
            if (sendto(sock,buf,sizeof(buf),0,(struct sockaddr *)&listen_dest,sizeof(listen_dest))==SOCKET_ERROR) {
              logfunc(applog,0,1,errno,1,"Error sending ANNOUNCE");
              break;
            } 
            if (packet_wait) usleep(packet_wait);
            bzero(info->addr_list,MAXDEST*sizeof(struct in_addr));
            j=0;
            confcnt++;
          }
        }
      } else {
        /* send open group ANNOUNCE */
        logfunc(applog,0,0,0,0,"Sending ANNOUNCE %d\n",attempt);
        if (sendto(sock,buf,sizeof(buf),0,(struct sockaddr *)&listen_dest,sizeof(listen_dest))==SOCKET_ERROR) {
          logfunc(applog,0,1,errno,1,"Error sending ANNOUNCE");
          attempt++;
          continue;
        }
      }
      resend=0;
    }
    FD_ZERO(&fdin);
    FD_SET(sock,&fdin);
    tv=tv_sav;
    if ((rval=select(FD_SETSIZE-1,&fdin,NULL,NULL,&tv))==SOCKET_ERROR) {
      logfunc(applog,0,1,errno,1,"Select failed");
      continue;
    }
    if (rval==0) {
      attempt++;
      resend=1;
    } else if (FD_ISSET(sock,&fdin)) {
      addr_len=sizeof(sin);
      bzero(buf,sizeof(buf));
      if ((rval=recvfrom(sock,buf,sizeof(buf),0,(struct sockaddr *)&sin,&addr_len))==SOCKET_ERROR) {
        logfunc(applog,0,1,errno,1,"Error receiving REGISTER");
        continue;
      } else if ((rval!=PACKETSIZE)&&(rval-sizeof(struct uftp_h)!=ntohl(header->blsize))) {
        logfunc(applog,0,1,0,0,"Invalid packet size %d for REGISTER",rval);
        continue;
      }
      head_ntoh(header);
      if (header->uftp_id!=UFTP_ID) {
        logfunc(applog,0,0,0,0,"Invalid REGISTER message: not uftp packet\n");
        continue;
      }
      if (header->tx_id!=txID) {
        logfunc(applog,0,0,0,0,"Invalid txID: %08X, should be %08X\n",header->tx_id,txID);
        abort_msg("Invalid txID",&sin);
        continue;
      }
      if (header->func==ABORT) {
        i=0;
        while (destlist[i].addr.s_addr!=sin.sin_addr.s_addr) {
          if (destlist[i].addr.s_addr==0)
            break;
          i++;
        }
        if (destlist[i].addr.s_addr==0) {
          logfunc(applog,0,0,0,0,"Transfer aborted by %s: %s\n",inet_ntoa(sin.sin_addr),data);
        } else {
          logfunc(applog,0,0,0,0,"Transfer aborted by %s: %s\n",destlist[i].name,data);
        }
        continue;
      } else if (header->func!=REGISTER) {
        logfunc(applog,0,0,0,0,"Invalid REGISTER message: wrong func: %s\n",strfunc(header->func));
        abort_msg("Invalid function",&sin);
        continue;
      }
      if (isclosed) {
        gotall=1;
        found=0;
        for (i=0;i<destcount;i++) {
          if (sin.sin_addr.s_addr==destlist[i].addr.s_addr) {
            if (destlist[i].done==-3) {
              logfunc(applog,0,0,0,0,"Received REGISTER from %s\n",destlist[i].name);
              destlist[i].done=0;
            } else {
              logfunc(applog,0,0,0,0,"Received REGISTER+ from %s\n",destlist[i].name);
            }
            destlist[i].naks=0;
            found=1;
          } 
          gotall=(gotall&&(destlist[i].done==0));
        }
        if (!found) {
          logfunc(applog,0,0,0,0,"Host %s not in host list\n",inet_ntoa(sin.sin_addr));
          abort_msg("Not in host list",&sin);
          continue;
        }
        if (gotall) {
          logfunc(applog,0,0,0,0,"Late registers:\n");
          polltime=latency[latency_lev][1];
          tv.tv_sec=polltime/1000;
          tv.tv_usec=(polltime%1000)*1000;
          tv_sav=tv;
          send_regconf(attempt+1);
          attempt=5;
        }
      } else {
        for (i=0;i<destcount;i++) {
          if (sin.sin_addr.s_addr==destlist[i].addr.s_addr) {
            logfunc(applog,0,0,0,0,"Received REGISTER+ from %s\n",destlist[i].name);
            i=-1;
            break;
          } 
        }
        if (i!=-1) {
          if (destcount==MAXOPENDEST) {
            logfunc(applog,0,0,0,0,"Rejecting REGISTER from %s: max destinations exceeded\n",inet_ntoa(sin.sin_addr));
            abort_msg("Max destinations exceeded",&sin);
          } else {
            if (!noname&&(hp=gethostbyaddr((char *)&sin.sin_addr,sizeof(struct in_addr),AF_INET))) {
              strncpy(destlist[destcount].name,hp->h_name,sizeof(destlist[0].name)-1);
            } else {
              strncpy(destlist[destcount].name,inet_ntoa(sin.sin_addr),sizeof(destlist[0].name)-1);
            }
            destlist[destcount].name[sizeof(destlist[0].name)-1]='\x0';
            destlist[destcount].addr=sin.sin_addr;
            destlist[destcount].done=0;
            logfunc(applog,0,0,0,0,"Received REGISTER from %s\n",destlist[destcount++].name);
          }
        }
      }
    } else {
      logfunc(applog,0,0,0,0,"unknown poll error\n");
    }
  }
  gotall=1;
  for (i=0;i<destcount;i++) {
    gotall=(gotall&&destlist[i].done);
  }
  if ((attempt>5)&&(gotall)) {
    logfunc(applog,0,0,0,0,"Announce timed out\n");
    exit(0);
  }
  if (!isclosed) {
    send_regconf(attempt);
  }
}

int getnaks(int pass, int section)
{
  fd_set fdin;
  struct timeval tv, tv_sav;
  struct sockaddr_in sin;
  char buf[PACKETSIZE],*list;
  struct uftp_h *header;
  struct statusinfo *sinfo;
  socklen_t addr_len;
  int attempt,rval,i,j,l_index,n_index,donecnt,senderr;
  int found,gotall,resend,numnaks,polltime;
  char *gotdest;
  int sectionoffset,blocksthissec;

  gotdest=malloc(MAXOPENDEST);
  sectionoffset=(BLOCKSIZE*8*(section-1));
  blocksthissec=((section<numsections)?(BLOCKSIZE*8):(numblocks%(BLOCKSIZE*8)));
  if (numsections && !blocksthissec) blocksthissec=BLOCKSIZE*8;
  header=(struct uftp_h *)&(buf[0]);
  sinfo=(struct statusinfo *)&(buf[sizeof(struct uftp_h)]);
  list=&(buf[sizeof(struct uftp_h)]);
  gotall=0;
  for (i=0;i<destcount;i++) {
    gotdest[i]=destlist[i].done;
  }
  for (i=0;i<blocksthissec;i++) {
    n_index=i+sectionoffset;
    naklist[n_index]=0;
  }
  polltime=latency[latency_lev][2];
  tv.tv_sec=polltime/1000;
  tv.tv_usec=(polltime%1000)*1000;
  tv_sav=tv;
  attempt=1;
  resend=1;
  if (packet_wait) usleep(packet_wait);
  while ((attempt<=3)&&(!gotall)) {
    if (resend) {
      send_doneconf(attempt);
      bzero(buf,sizeof(buf));
      header->uftp_id=UFTP_ID;
      header->func=DONE;
      header->tx_id=txID;
      header->pass=pass;
      header->section_num=section;
      head_hton(header);
      donecnt=0;
      senderr=0;
      for (i=0,j=0;i<destcount;i++) {
        if (!gotdest[i]) {
          sinfo->addr_list[j++]=destlist[i].addr;
        }
        if ((j>=MAXINFODEST)||(i==destcount-1)) {
          logfunc(applog,0,0,0,0,"sending DONE %d%c for pass %d section %d\n",attempt,'a'+donecnt,pass,section);
          if (sendto(sock,buf,sizeof(buf),0,(struct sockaddr *)&receive_dest,sizeof(receive_dest))==SOCKET_ERROR) {
            logfunc(applog,0,1,errno,1,"Error sending DONE");
            attempt++;
            senderr=1;
            break;
          }
          bzero(sinfo->addr_list,MAXINFODEST*sizeof(struct in_addr));
          j=0;
          donecnt++;
        }
      }
      if (senderr) continue;
      resend=0;
    }
    FD_ZERO(&fdin);
    FD_SET(sock,&fdin);
    tv=tv_sav;
    if ((rval=select(FD_SETSIZE-1,&fdin,NULL,NULL,&tv))==SOCKET_ERROR) {
      logfunc(applog,0,1,errno,1,"Select failed");
      attempt++;
      continue;
    }
    if (rval==0) {
      attempt++;
      resend=1;
    } else if (FD_ISSET(sock,&fdin)) {
      addr_len=sizeof(sin);
      bzero(buf,sizeof(buf));
      if ((rval=recvfrom(sock,buf,sizeof(buf),0,(struct sockaddr *)&sin,&addr_len))==SOCKET_ERROR) {
        logfunc(applog,0,1,errno,1,"Error receiving STATUS");
        continue;
      } else if ((rval!=PACKETSIZE)&&(rval-sizeof(struct uftp_h)!=ntohl(header->blsize))) {
        logfunc(applog,0,1,0,0,"Invalid packet size %d for STATUS",rval);
        continue;
      }
      head_ntoh(header);
      if (header->uftp_id!=UFTP_ID) {
        logfunc(applog,0,0,0,0,"Invalid STATUS message: not uftp packet\n");
        continue;
      }
      if (header->tx_id!=txID) {
        logfunc(applog,0,0,0,0,"Invalid txID: %08X, should be %08X\n",header->tx_id,txID);
        continue;
      }
      if (header->func==ABORT) {
        for (i=0;i<destcount;i++) {
          if (destlist[i].addr.s_addr==sin.sin_addr.s_addr) {
            logfunc(applog,0,0,0,0,"Transfer aborted by %s: %s\n",destlist[i].name,list);
            break;
          }
        }
        if (i==destcount) {
          logfunc(applog,0,0,0,0,"Transfer aborted by unknown (%s): %s\n",inet_ntoa(sin.sin_addr),list);
          continue;
        }
        gotdest[i]=-1;
        destlist[i].done=-1;
        gotall=1;
        for (i=0;i<destcount;i++)
          gotall=(gotall&&gotdest[i]);
        continue;
      } else if (header->func!=STATUS) {
        logfunc(applog,0,0,0,0,"Invalid STATUS message: wrong func: %s\n",strfunc(header->func));
        continue;
      } else if ((rval!=PACKETSIZE)&&(header->nak_count>0)) {
        logfunc(applog,0,1,0,0,"Invalid packet size %d for STATUS",rval);
        continue;
      }
      found=0;
      for (i=0;i<destcount;i++) {
        if (destlist[i].addr.s_addr==sin.sin_addr.s_addr) {
          if (header->nak_count==-1) {
            if (destlist[i].done!=1) {
              logfunc(applog,0,0,0,0,"Got completion from %s\n",destlist[i].name);
            } else {
              logfunc(applog,0,0,0,0,"Got completion+ from %s\n",destlist[i].name);
            }
            destlist[i].done=1;
            gotdest[i]=1;
            gettimestamp(destlist[i].time);
            found=-1;
          } else if (gotdest[i]==1) {
            found=-1;
          } else {
            logfunc(applog,0,0,0,0,"Got %d NAKs for pass %d section %d from %s\n",header->nak_count,header->pass,header->section_num,destlist[i].name);
            if ((header->pass==pass)&&(header->section_num==section)) {
              gotdest[i]=1;
              destlist[i].naks+=header->nak_count;
            }
            found=1;
          }
          break;
        }
      }
      if (found==0) {
        logfunc(applog,0,0,0,0,"Unknown address: %s\n",inet_ntoa(sin.sin_addr));
      } else if ((found!=-1)&&(header->pass==pass)&&(header->section_num==section)&&(header->nak_count>0)) {
        for (i=0;i<blocksthissec;i++) {
          n_index=i+sectionoffset;
          l_index=i;
          if ((list[l_index>>3] & (1<<(l_index&7)))!=0) {
            if (verbose>=2)
              logfunc(applog,0,0,0,0,"Got NAK for %d\n",n_index);
            naklist[n_index]=1;
          }
        }
      }
      gotall=1;
      for (i=0;i<destcount;i++)
        gotall=(gotall&&destlist[i].done);
      if (gotall) {
        logfunc(applog,0,0,0,0,"Late completions:\n");
        polltime=latency[latency_lev][3];
        tv.tv_sec=polltime/1000;
        tv.tv_usec=(polltime%1000)*1000;
        tv_sav=tv;
        send_doneconf(attempt+1);
        attempt=3;
        gotall=0;
        continue;
      }
      gotall=1;
      for (i=0;i<destcount;i++)
        gotall=(gotall&&gotdest[i]);
    } else {
      logfunc(applog,0,0,0,0,"unknown poll error\n");
      attempt++;
    }
  }
  if (attempt>3) {
    for (i=0;i<destcount;i++) {
      if (!gotdest[i]) {
        logfunc(applog,0,0,0,0,"Couldn't get STATUS from %s\n",destlist[i].name);
        destlist[i].done=-2;
      }
    }
  } else {
    send_doneconf(attempt+1);
  }
  numnaks=0;
  for (i=0;i<blocksthissec;i++) {
    n_index=i+sectionoffset;
    if (naklist[n_index]) {
      numnaks++;
    }
  }
  free(gotdest);
  return numnaks;
}

void xfer()
{
  char buf[PACKETSIZE],*data;
  struct uftp_h *header;
  int32_t i,numnaks;
  int numbytes,pass,weight_time,max_time,aborted,naksthissec;
  timestamp t1,t2,w1,w2;
  double t3;
  int32_t avg,count,overage,tdiff;
  f_offset_t offset,prev_offset;

  bzero(buf,sizeof(buf));
  bzero(naklist,numblocks);
  header=(struct uftp_h *)&(buf[0]);
  data=&(buf[sizeof(struct uftp_h)]);
  if ((file=open(filename,OPENREAD,0))==-1) {
    logfunc(applog,0,1,errno,0,"Error opening file");
    exit(1);
  }
  if (rate==-1) {
    /* use 100Mbps for purpose of calculating weight */
    weight_time=(int)floor(((double)weight/100)*((double)filesize/(100000/8)/1024));
    max_time=(weight_time>min_time)?weight_time:min_time;
    logfunc(applog,0,0,0,0,"Transferring at full speed!\n");
  } else {
    weight_time=(int)floor(((double)weight/100)*((double)filesize/(rate/8)/1024));
    max_time=(weight_time>min_time)?weight_time:min_time;
    logfunc(applog,0,0,0,0,"Transfer rate: %d Kbps (%d KB/s)\n",rate,rate/8);
    logfunc(applog,0,0,0,0,"Wait between packets: %d %cs\n",packet_wait,MICRO);
  }
  logfunc(applog,0,0,0,0,"Maximum file transfer time: %d seconds\n",max_time);
  aborted=0;
  pass=1;
  if (packet_wait) usleep(packet_wait);
  gettimestamp(t1);
  do {
    if (aborted) break;
    avg=0; count=0;
    numnaks=0;
    overage=0;
    header->uftp_id=UFTP_ID;
    header->func=FILESEG;
    header->tx_id=txID;
    header->seq_num=-1;
    header->section_num=1;
    gettimestamp(w1);
    logfunc(applog,0,0,0,0,"Sending file...pass %d\n",pass);
    if (pass==1) {
      lseek_func(file,0,SEEK_SET);
      while ((numbytes=read(file,data,BLOCKSIZE))!=0) {
        if (diff_sec(w1,t1)>max_time) {
          logfunc(applog,0,0,0,0,"Max file transfer time exceeded\n");
          abort_msg("Max file transfer time exceeded",&receive_dest);
          aborted=1;
          for (i=0;i<destcount;i++) {
            if (!destlist[i].done) destlist[i].done=-1;
          }
          break;
        }
        header->blsize=numbytes;
        header->pass=pass;
        header->seq_num++;
        if (numbytes==-1) {
          logfunc(applog,0,0,0,0,"read failed\n");
          continue;
        }
        if (packet_wait>overage) usleep(packet_wait-overage);
        gettimestamp(w2);
        tdiff=diff_usec(w2,w1);
        avg+=tdiff;
        count++;
        if (packet_wait) overage+=tdiff-packet_wait;
        w1=w2;
        head_hton(header);
        if (sendto(sock,buf,sizeof(buf),0,(struct sockaddr *)&receive_dest,sizeof(receive_dest))==SOCKET_ERROR) {
          logfunc(applog,0,1,errno,1,"Error sending FILESEG");
          continue;
        }
        head_ntoh(header);
        if ((header->seq_num%(BLOCKSIZE*8)==(BLOCKSIZE*8)-1)&&(header->seq_num!=numblocks-1)) {
          numnaks+=getnaks(header->pass=pass,header->section_num);
          gettimestamp(w1);
          header->section_num++;
        }
      }
    } else {
      naksthissec=0;
      prev_offset=0;
      lseek_func(file,0,SEEK_SET);
      for (i=0;i<numblocks;i++) {
        if (naklist[i]) {
          naksthissec=1;
          if (diff_sec(w1,t1)>max_time) {
            logfunc(applog,0,0,0,0,"Max file transfer time exceeded\n");
            abort_msg("Max file transfer time exceeded",&receive_dest);
            aborted=1;
            for (i=0;i<destcount;i++) {
              if (!destlist[i].done) destlist[i].done=-1;
            }
            break;
          }
          if (verbose>=2)
            logfunc(applog,0,0,0,0,"Resending %d\n",i);
          if ((offset=lseek_func(file,((f_offset_t)i*BLOCKSIZE)-prev_offset,SEEK_CUR))==-1) {
            logfunc(applog,0,1,errno,0,"lseek failed for file");
            continue;
          }
          if (offset!=(f_offset_t)i*BLOCKSIZE) {
            logfunc(applog,0,0,0,0,"sizeof f_offset_t = %d\n",sizeof(f_offset_t));
            if (sizeof(f_offset_t)>4)
              logfunc(applog,0,0,0,0,"block %d: offset is %s, should be %s\n",i,printll(offset),printll((f_offset_t)i*BLOCKSIZE));
            else
              logfunc(applog,0,0,0,0,"block %d: offset is %ld, should be %ld\n",i,offset,(f_offset_t)i*BLOCKSIZE);
            continue;
          }
          if ((numbytes=read(file,data,BLOCKSIZE))==-1) {
            logfunc(applog,0,0,0,0,"read failed\n");
            continue;
          }
          prev_offset=offset+numbytes;
          header->blsize=numbytes;
          header->pass=pass;
          header->seq_num=i;
          if (packet_wait>overage) usleep(packet_wait-overage);
          gettimestamp(w2);
          tdiff=diff_usec(w2,w1);
          avg+=tdiff;
          count++;
          if (packet_wait) overage+=tdiff-packet_wait;
          w1=w2;
          head_hton(header);
          if (sendto(sock,buf,sizeof(buf),0,(struct sockaddr *)&receive_dest,sizeof(receive_dest))==SOCKET_ERROR) {
            logfunc(applog,0,1,errno,1,"Error sending FILESEG");
            continue;
          }
          head_ntoh(header);
        }
        if ((i%(BLOCKSIZE*8)==(BLOCKSIZE*8)-1)&&(i!=numblocks-1)) {
          if (naksthissec) {
            numnaks+=getnaks(header->pass=pass,header->section_num);
          }
          naksthissec=0;
          gettimestamp(w1);
          header->section_num++;
        }
      }
    }
    if (!aborted)
      numnaks+=getnaks(header->pass=pass,header->section_num);
    gettimestamp(w1);
    logfunc(applog,0,0,0,0,"average wait time = %.2f %cs\n",(count==0)?0:(float)avg/count,MICRO);
    logfunc(applog,0,0,0,0,"Received %d distinct NAKs for pass %d\n",numnaks,pass);
    pass++;
  } while (numnaks);
  close(file);
  file=-1;
  for (t2=t1,i=0;i<destcount;i++) {
    if ((destlist[i].done==1)&&(diff_usec(destlist[i].time,t2)>0))
      t2=destlist[i].time;
  }
  t3=(double)diff_usec(t2,t1)/1000000;
  logfunc(applog,0,0,0,0,"total elapsed time: %.3f seconds\n",t3);
  logfunc(applog,0,0,0,0,"throughput: %.2f KB/s\n",(t3!=0)?(filesize/t3/1024):0);
  logfunc(applog,0,0,0,0,"Transfer status:\n"); 
  for (i=0;i<destcount;i++) {
    logfunc(applog,0,0,0,0,"host: %-15s  status: ",destlist[i].name);
    switch (destlist[i].done) {
    case -3:
      logfunc(applog,0,0,0,0,"Mute\n");
      break;
    case -2:
      logfunc(applog,0,0,0,0,"Lost connection\n");
      break;
    case -1:
      logfunc(applog,0,0,0,0,"Aborted\n");
      break;
    case 1:
      t3=(double)diff_usec(destlist[i].time,t1)/1000000;
      logfunc(applog,0,0,0,0,"Completed   time: %7.3f seconds    NAKs: %d\n",t3,destlist[i].naks);
      break;
    default:
      logfunc(applog,0,0,0,0,"Unknown  code: %d\n",destlist[i].done);
      break;
    }
  }
}

int main(int argc, char *argv[])
{
  struct iflist ifl[MAX_INTERFACES];
  struct sockaddr_in sin;
  stat_struct statbuf;
  struct hostent *hp;
  struct in_addr *addr,out_addr;
  char pub_multi[20],priv_multi[20],tmp_multi[20],*p;
  char logfile[MAXPATHNAME];
  uint16_t port;
  time_t t;
  int32_t buffer;
  int c,i,ifl_len,foundif;
  unsigned char ttl;

#ifdef WINDOWS
  struct WSAData data;
  if (WSAStartup(2,&data)) {
    logfunc(stderr, 0, 1, 1, 1, "Error in WSAStartup");
    exit(1);
  }
#endif

  ifl_len=sizeof(ifl)/sizeof(struct iflist);
  getiflist(ifl,&ifl_len);
  for (c=0;c<MAXOPENDEST;c++)
    bzero(&destlist[c],sizeof(struct destinfo_t));
  file=-1;
  destcount=0;
  naklist=NULL;
  port=DEF_PORT;
  rate=DEF_RATE;
  weight=DEF_WEIGHT;
  min_time=DEF_MIN_TIME;
  latency_lev=0;
  out_addr.s_addr=htonl(INADDR_ANY);
  ttl=1;
  verbose=0;
  unicast=0;
  noname=0;
  client_wait=0;
  minimal_resp=0;
  buffer=0;
  strncpy(pub_multi,DEF_PUB_MULTI,sizeof(pub_multi)-1);
  pub_multi[sizeof(pub_multi)-1]='\x0';
  strncpy(priv_multi,DEF_PRIV_MULTI,sizeof(priv_multi)-1);
  priv_multi[sizeof(priv_multi)-1]='\x0';
  strncpy(logfile,"",sizeof(logfile)-1);
  logfile[sizeof(logfile)-1]='\x0';
  while ((c=getopt(argc,argv,"Uv:R:W:m:nzl:c:t:I:p:L:H:B:M:P:")) != EOF) {
    switch (c) {
    case 'U':
      unicast=1;
      break;
    case 'v':
      verbose=atoi(optarg);
      break;
    case 'R':
      if ((atoi(optarg)<=0)&&(atoi(optarg)!=-1))
        fprintf(stderr,"Invalid rate\n");
      else
        rate=atoi(optarg);
      break;
    case 'W':
      if ((atoi(optarg)<110)||(atoi(optarg)>10000))
        fprintf(stderr,"Invalid weight\n");
      else
        weight=atoi(optarg);
      break;
    case 'm':
      if ((atoi(optarg)<=0)||(atoi(optarg)>3600))
        fprintf(stderr,"Invalid min time\n");
      else
        min_time=atoi(optarg);
      break;
    case 'n':
      noname=1;
      break;
    case 'z':
      minimal_resp=1;
      break;
    case 'l':
      if ((atoi(optarg)<=0)||(atoi(optarg)>latency_lev_count))
        fprintf(stderr,"Invalid latency level\n");
      else
        latency_lev=atoi(optarg);
      break;
    case 'c':
      if ((atoi(optarg)<=0)||(atoi(optarg)>60000))
        fprintf(stderr,"Invalid client wait\n");
      else
        client_wait=atoi(optarg);
      break;
    case 't':
      if ((atoi(optarg)<=0)||(atoi(optarg)>255))
        fprintf(stderr,"Invalid ttl\n");
      else
        ttl=atoi(optarg);
      break;
    case 'I':
      for (i=0,foundif=0;(i<ifl_len)&&(!foundif);i++) {
        if (!strcmp(optarg,ifl[i].name)) {
          foundif=1;
        }
      }
      if (foundif) {
        out_addr.s_addr=ifl[i-1].addr.s_addr;
        break;
      }
      if (inet_addr(optarg)==INADDR_NONE) {
        if ((hp=gethostbyname(optarg))==NULL) {
          fprintf(stderr,"Invalid host name: %s\n",optarg);
        } else {
          addr=(struct in_addr *)hp->h_addr_list[0];
          out_addr.s_addr=addr->s_addr;
        }
      } else {
        out_addr.s_addr=inet_addr(optarg);
      }
      break;
    case 'p':
      port=atoi(optarg);
      if (port==0) {
        fprintf(stderr,"Invalid port\n");
        exit(1);
      }
      break;
    case 'L':
      strncpy(logfile,optarg,sizeof(logfile)-1);
      logfile[sizeof(logfile)-1]='\x0';
      break;
    case 'H':
      p=(char *)strtok(optarg,",");
      while (p!=NULL) {
        if (destcount==MAXOPENDEST) {
          fprintf(stderr,"Exceeded maximum destination count of %d for closed group membership\n",MAXOPENDEST);
          exit(1);
        }
        if (inet_addr(p)==INADDR_NONE) {
          if (noname) {
            fprintf(stderr,"Invalid IP: %s\n",p);
            exit(1);
          } else if ((hp=gethostbyname(p))==NULL) {
            fprintf(stderr,"Invalid host name: %s\n",p);
            exit(1);
          } else {
            addr=(struct in_addr *)hp->h_addr_list[0];
            destlist[destcount].addr=*addr;
            strncpy(destlist[destcount].name,p,sizeof(destlist[0].name));
            destlist[destcount].name[sizeof(destlist[0].name)-1]='\x0';
            destcount++;
          }
        } else {
          destlist[destcount].addr.s_addr=inet_addr(p);
          strncpy(destlist[destcount].name,p,sizeof(destlist[0].name));
          destlist[destcount].name[sizeof(destlist[0].name)-1]='\x0';
          destcount++;
        }
        p=(char *)strtok(NULL,",");
      }
      break;
    case 'B':
      buffer=atoi(optarg);
      if ((buffer<65536)||(buffer>104857600)) {
        fprintf(stderr,"Invalid buffer size\n");
        exit(1);
      }
      break;
    case 'M':
      strncpy(pub_multi,optarg,sizeof(pub_multi)-1);
      pub_multi[sizeof(pub_multi)-1]='\x0';
      break;
    case 'P':
      strncpy(priv_multi,optarg,sizeof(priv_multi)-1);
      priv_multi[sizeof(priv_multi)-1]='\x0';
      break;
    case '?':
      fprintf(stderr,USAGE);
      exit(1);
    }
  }  
  argc-=optind;
  argv+=optind;
  if (argc!=1) {
    fprintf(stderr,USAGE);
    exit(1);
  }
  if (unicast&&(destcount!=1)) {
    fprintf(stderr,"Must specify exactly one host for unicast\n");
    exit(1);
  }
  if (strcmp(logfile,"")) {
#ifdef WINDOWS
    if ((applog=fopen(logfile,"at"))==NULL) {
      perror("Can't open log file");
      exit(1);
    }
#else
    int fd;
    if ((fd=open(logfile,O_WRONLY|O_APPEND|O_CREAT,0644))==-1) {
      perror("Can't open log file");
      exit(1);
    }
    close(2);
    dup(fd);
    close(fd);
    applog=stderr;
#endif
  } else {
    applog=stderr;
  }
  t=time(NULL);
  logfunc(applog,0,0,0,0,"\n%s\n",VERSION);
  logfunc(applog,0,0,0,0,"Starting at %s",ctime(&t));
  if ((sock=socket(AF_INET,SOCK_DGRAM,0))==INVALID_SOCKET) {
    logfunc(applog,0,1,errno,1,"Error creating socket");
    exit(1);
  }
  atexit(cleanup);
  sin.sin_family=AF_INET;
  sin.sin_addr.s_addr=htonl(INADDR_ANY);
  sin.sin_port=0;
  if (bind(sock,(struct sockaddr *)&sin,sizeof(sin))==SOCKET_ERROR) {
    logfunc(applog,0,1,errno,1,"Error binding socket");
    exit(1);
  }
  if (buffer) {
    if (setsockopt(sock,SOL_SOCKET,SO_RCVBUF,(char *)&buffer,sizeof(buffer))==SOCKET_ERROR) {
      logfunc(applog,0,1,errno,1,"Error setting receive buffer size");
      exit(1);
    }
  } else {
    buffer=262144;
    if (setsockopt(sock,SOL_SOCKET,SO_RCVBUF,(char *)&buffer,sizeof(buffer))==SOCKET_ERROR) {
      buffer=233016;
      if (setsockopt(sock,SOL_SOCKET,SO_RCVBUF,(char *)&buffer,sizeof(buffer))==SOCKET_ERROR) {
        logfunc(applog,0,1,errno,1,"Error setting receive buffer size");
        exit(1);
      }
    }
  }
  if (setsockopt(sock,IPPROTO_IP,IP_MULTICAST_TTL,&ttl,sizeof(ttl))==SOCKET_ERROR) {
    logfunc(applog,0,1,errno,1,"Error setting ttl");
    closesocket(sock);
    exit(1);
  }
  if (setsockopt(sock,IPPROTO_IP,IP_MULTICAST_IF,(char *)&out_addr,sizeof(out_addr))==SOCKET_ERROR) {
    logfunc(applog,0,1,errno,1,"Error setting outgoing interface");
    closesocket(sock);
    exit(1);
  }
  listen_dest.sin_family=AF_INET;
  listen_dest.sin_addr.s_addr=inet_addr(pub_multi);
  listen_dest.sin_port=htons(port);
  srand(getpid());
  while ((p=(char *)strchr(priv_multi,'x'))!=NULL) {
    bzero(tmp_multi,sizeof(tmp_multi));
    strncpy(tmp_multi,priv_multi,p-priv_multi);
    sprintf(tmp_multi,"%s%d%s",tmp_multi,rand()&255,p+1);
    strcpy(priv_multi,tmp_multi);
  }
  receive_dest.sin_family=AF_INET;
  receive_dest.sin_addr.s_addr=inet_addr(priv_multi);
  receive_dest.sin_port=htons(port);
  if (unicast) {
    listen_dest.sin_addr=destlist[0].addr;
    receive_dest.sin_addr=destlist[0].addr;
  }
  strncpy(filename,argv[0],sizeof(filename)-1);
  filename[sizeof(filename)-1]='\x0';
  if (stat_func(filename,&statbuf)==-1) {
    logfunc(applog,0,1,errno,0,"Error getting file status");
    exit(1);
  }
  srand((unsigned)time(NULL));
  txID=rand()&0xFFFF;
  txID<<=16;
  srand(getpid());
  txID|=rand()&0xFFFF;
  filesize=statbuf.st_size;
  numblocks=(int32_t)((filesize/BLOCKSIZE)+(filesize%BLOCKSIZE?1:0));
  numsections=(numblocks/(BLOCKSIZE*8))+(numblocks%(BLOCKSIZE*8)?1:0);
  naklist=(char *)malloc(numblocks);
  logfunc(applog,0,0,0,0,"Name of file to send: %s\n",filename);
  if (sizeof(f_offset_t)>4) 
    logfunc(applog,0,0,0,0,"Bytes: %s  Blocks: %d  Sections: %d\n",printll(filesize),numblocks,numsections);
  else
    logfunc(applog,0,0,0,0,"Bytes: %d  Blocks: %d  Sections: %d\n",filesize,numblocks,numsections);
  logfunc(applog,0,0,0,0,"Using private multicast address %s  id %08X\n",inet_ntoa(receive_dest.sin_addr),txID);
  if (rate==-1) {
    packet_wait=0;
  } else {
    packet_wait=(int32_t)(1000000*PACKETSIZE/((float)rate*1024/8));
  }
  announce();
  xfer();
  t=time(NULL);
  logfunc(applog,0,0,0,0,"uftp: Finishing at %s",ctime(&t));
  return 0;
}
