/*
 *  UFTP - UDP based FTP with multicast
 *
 *  Copyright (C) 2001-2010   Dennis A. Bush, Jr.   bush@tcnj.edu
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 3 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License along
 *  with this program; if not, write to the Free Software Foundation, Inc.,
 *  51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
 */

#include <sys/types.h>
#include <sys/stat.h>
#include <stdio.h>
#include <fcntl.h>
#include <time.h>
#include <math.h>
#include <string.h>
#include <stdlib.h>
#include <signal.h>
#include <sys/timeb.h>
#include <stdarg.h>
#include <errno.h>

#if ( defined WIN16 || defined WIN32 || defined WIN64 )

#define WINDOWS
#include <winsock2.h>
#include <ws2tcpip.h>
#include <winbase.h>
#include <io.h>
#include <process.h>
#include <winnt.h>
#include "win_func.h"

typedef unsigned __int8 uint8_t;
typedef __int8 int8_t;
typedef unsigned __int16 uint16_t;
typedef __int16 int16_t;
typedef unsigned __int32 uint32_t;
typedef __int32 int32_t;
typedef unsigned __int64 uint64_t;
typedef __int64 int64_t;

#define link(src,dest) rename(src,dest)
#define bzero(buf,size) memset(buf,0,size)
#define usleep(t) Sleep((t)/1000)
#define isfullpath(str) (str[1]==':')
#define PATH_SEP '\\'

typedef int64_t f_offset_t;
typedef struct __stat64 stat_struct;
#define stat_func(p,s) _stat64(p,s)
#define lseek_func(f,o,l) _lseeki64(f,o,l)
#define snprintf _snprintf
#define open _open
#define close _close
#define read _read
#define write _write
#define unlink _unlink
#define getpid _getpid

typedef LARGE_INTEGER timestamp;
#define gettimestamp(t) QueryPerformanceCounter(&t)

#define OPENREAD (O_RDONLY|O_BINARY)
#define OPENWRITE (O_WRONLY|O_BINARY)

#define MICRO '\xe6'

#else  /*if WINDOWS*/

#include <netinet/in.h>
#include <sys/socket.h>
#include <poll.h>
#include <unistd.h>
#include <netdb.h>
#include <strings.h>
#include <arpa/inet.h>
#include <sys/time.h>
#include <libgen.h>
#include <net/if.h>
#include <sys/ioctl.h>

#define closesocket(s) close(s)

#ifdef VMS
pid_t GENERIC_SETSID(void);
#define setsid GENERIC_SETSID
#define fork vfork
typedef unsigned int socklen_t;
#define isfullpath(str) (1)
#define PATH_SEP ':'
#define open(a,b,c) open(a,b,c,"ctx=stm")
#else
#define isfullpath(str) (str[0]=='/')
#define PATH_SEP '/'
#endif

typedef int64_t f_offset_t;
typedef struct stat stat_struct;
#define stat_func(p,s) stat(p,s)
#define lseek_func(f,o,l) lseek(f,o,l)

typedef struct timeval timestamp;
#define gettimestamp(t) gettimeofday(&t,NULL)

typedef int SOCKET;
#define INVALID_SOCKET -1
#define SOCKET_ERROR -1
#ifndef INADDR_NONE
#define INADDR_NONE -1
#endif
#define OPENREAD (O_RDONLY)
#define OPENWRITE (O_WRONLY)

#define MICRO '\xb5'

#endif /*if WINDOWS*/

#define UFTP_ID		0x69
#define ANNOUNCE	1
#define REGISTER	2
#define REG_CONF	3
#define FILESEG		4
#define DONE		5
#define STATUS		6
#define DONE_CONF	7
#define ABORT		99

#define BLOCKSIZE 1400
#define PACKETSIZE (BLOCKSIZE+sizeof(struct uftp_h))
#define MAXDEST 100
#define MAXINFODEST 300
#define MAXOPENDEST 10000
#define MAXFILENAME 100
#define MAXDIRNAME 200
#define MAXPATHNAME 300

#define VERSION "UFTP version 2.10.3  Copyright (C) 2001-2010  Dennis A. Bush"

struct uftp_h {
  int32_t uftp_id;
  int32_t func;
  uint32_t tx_id;
  int32_t seq_num;
  int32_t section_num;
  int32_t nak_count;
  int32_t blsize;
  int32_t pass;
};

/* As of 2.6.2 all code buffers for filenames are 100 bytes.
 * Leave it at 300 bytes here for backward compatibility of the protocol.
 */
struct fileinfo {
  uint32_t fsize;
  int32_t block_total;
  int32_t section_total;
  int32_t open;
  struct in_addr mcast;
  struct in_addr addr_list[MAXDEST];
  char name[300];
  uint16_t rxlatency1;
  uint16_t rxlatency2;
  uint32_t hilargefsize,lolargefsize;
  uint8_t minimal_resp;
};

/* As of 2.8 this structure is used for the body of REG_CONF, DONE, and
 * DONE_CONF.  The address list is extended but left in the same place
 * relative to the original fileinfo for backward compatibility.
 */
struct statusinfo {
  uint32_t unused1;
  int32_t unused2;
  int32_t unused3;
  int32_t unused4;
  struct in_addr unused5;
  struct in_addr addr_list[MAXINFODEST];
};

const char *strfunc(int func);
void head_hton(struct uftp_h *header);
void head_ntoh(struct uftp_h *header);
void info_hton(struct fileinfo *info);
void info_ntoh(struct fileinfo *info);
void logfunc(FILE *logf, uint32_t id, int newline, int err, int sockerr, const char str[], ...);

struct iflist {
  char name[25];
  struct in_addr addr;
  int isloopback;
};

int32_t diff_sec(timestamp t2, timestamp t1);
int32_t diff_usec(timestamp t2, timestamp t1);
timestamp add_msec_interval(timestamp t, int interval);
timestamp add_sec_interval(timestamp t, int interval);
int cmptimestamp(timestamp t1, timestamp t2);
const char *printll(int64_t n);
void getiflist(struct iflist *list, int *len);

