/*
 *  UFTP - UDP based FTP with multicast
 *
 *  Copyright (C) 2001-2010   Dennis A. Bush, Jr.   bush@tcnj.edu
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 3 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License along
 *  with this program; if not, write to the Free Software Foundation, Inc.,
 *  51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
 */

#include "uftp.h"

#ifdef WINDOWS

int32_t diff_sec(timestamp t2, timestamp t1)
{ 
  timestamp f;
  
  QueryPerformanceFrequency(&f);
  return (int32_t)((t2.QuadPart-t1.QuadPart)/f.QuadPart);
}

int32_t diff_usec(timestamp t2, timestamp t1)
{ 
  timestamp f;
  
  QueryPerformanceFrequency(&f);
  return (int32_t)(1000000*(t2.QuadPart-t1.QuadPart)/f.QuadPart);
}

timestamp add_msec_interval(timestamp t, int interval)
{
  timestamp f,new_interval,new_time;
  
  QueryPerformanceFrequency(&f);
  new_interval.QuadPart=(LONG64)interval*f.QuadPart/1000;
  new_time.QuadPart=t.QuadPart+new_interval.QuadPart;
  return new_time;
}

timestamp add_sec_interval(timestamp t, int interval)
{
  timestamp f,new_interval,new_time;
  
  QueryPerformanceFrequency(&f);
  new_interval.QuadPart=(LONG64)interval*f.QuadPart;
  new_time.QuadPart=t.QuadPart+new_interval.QuadPart;
  return new_time;
}

int cmptimestamp(timestamp t1, timestamp t2)
{
  if (t1.QuadPart>t2.QuadPart) {
    return 1;
  } else if (t1.QuadPart<t2.QuadPart) {
    return -1;
  } else {
    return 0;
  }
}

const char *printll(int64_t n)
{
  static char str[50];
  snprintf(str,sizeof(str)-1,"%I64d",n);
  str[sizeof(str)-1]='\x0';
  return &str[0];
}

void getiflist(struct iflist *list, int *len)
{
  SOCKET s;
  int rval,rsize,count,i;
  INTERFACE_INFO *iflist;

  if (*len<=0) return;
  count=*len;
  iflist=malloc(sizeof(INTERFACE_INFO)*count);

  if ((s=socket(AF_INET,SOCK_DGRAM,0))==INVALID_SOCKET) {
    logfunc(stderr, 0, 1, 1, 1, "Error creating socket for interface list");
    free(iflist);
    *len=0;
    return;
  }
  if ((rval=WSAIoctl(s,SIO_GET_INTERFACE_LIST,NULL,0,iflist,sizeof(INTERFACE_INFO)*count,&rsize,NULL,NULL))==SOCKET_ERROR) {
    logfunc(stderr, 0, 1, 1, 1, "Error getting interface list");
    closesocket(s);
    free(iflist);
    *len=0;
    return;
  }
  count=rsize/sizeof(INTERFACE_INFO);
  for (i=0,*len=0;i<count;i++) {
    if ((iflist[i].iiFlags&IFF_UP)!=0) {
      memset(&list[*len],0,sizeof(struct iflist));
      strcpy(list[*len].name,"");
      list[*len].addr.s_addr=iflist[i].iiAddress.AddressIn.sin_addr.s_addr;
      list[*len].isloopback=(iflist[i].iiFlags&IFF_LOOPBACK)!=0;
      (*len)++;
    }
  }
  free(iflist);
  closesocket(s);
}

#else  /*if WINDOWS*/

int32_t diff_sec(timestamp t2, timestamp t1)
{
  return t2.tv_sec-t1.tv_sec;
}

int32_t diff_usec(timestamp t2, timestamp t1)
{
  return (t2.tv_usec-t1.tv_usec)+1000000*(t2.tv_sec-t1.tv_sec);
}

timestamp add_msec_interval(timestamp t, int interval)
{
  timestamp new_time;
  
  new_time=t;
  new_time.tv_usec+=interval*1000;
  while (new_time.tv_usec>=1000000) {
    new_time.tv_usec-=1000000;
    new_time.tv_sec+=1;
  }
  return new_time;
}

timestamp add_sec_interval(timestamp t, int interval)
{
  timestamp new_time;
  
  new_time=t;
  new_time.tv_sec+=interval;
  return new_time;
}

int cmptimestamp(timestamp t1, timestamp t2)
{
  if (t1.tv_sec>t2.tv_sec) {
    return 1;
  } else if (t1.tv_sec<t2.tv_sec) {
    return -1;
  } else if (t1.tv_usec>t2.tv_usec) {
    return 1;
  } else if (t1.tv_usec<t2.tv_usec) {
    return -1;
  } else {
    return 0;
  }
}

const char *printll(int64_t n)
{
  static char str[50];
  snprintf(str,sizeof(str)-1,"%lld",n);
  str[sizeof(str)-1]='\x0';
  return &str[0];
}

#ifdef FreeBSD

#include <ifaddrs.h>

void getiflist(struct iflist *list, int *len)
{
  struct ifaddrs *ifa,*ifa_tmp;
  struct sockaddr_in *in;

  if (getifaddrs(&ifa)==-1) {
    perror("getifaddrs failed");
    *len=0;
    return;
  }
  ifa_tmp=ifa;
  *len=0;
  while (ifa_tmp) {
    in=(struct sockaddr_in *)ifa_tmp->ifa_addr;
    if (in&&(in->sin_family==AF_INET)&&((ifa_tmp->ifa_flags&IFF_UP)!=0)) {
      memset(&list[*len],0,sizeof(struct iflist));
      strncpy(list[*len].name,ifa_tmp->ifa_name,sizeof(list[*len].name)-1);
      list[*len].addr.s_addr=in->sin_addr.s_addr;
      list[*len].isloopback=(ifa_tmp->ifa_flags&IFF_LOOPBACK)!=0;
      (*len)++;
    }
    ifa_tmp=ifa_tmp->ifa_next;
  }
  freeifaddrs(ifa);
}

#else

void getiflist(struct iflist *list, int *len)
{
  int s,i,count;
  struct ifconf ifc;
  struct ifreq *ifr;
  struct ifreq ifr_tmp;
  struct sockaddr_in *in;

  if (*len<=0) return;
  count=*len;
  ifr=malloc(sizeof(struct ifreq)*count);
  ifc.ifc_len=sizeof(struct ifreq)*count;
  ifc.ifc_req=ifr;

  if ((s=socket(AF_INET,SOCK_DGRAM,0))==-1) {
    perror("Error creating socket for interface list");
    free(ifr);
    *len=0;
    return;
  }
  if (ioctl(s,SIOCGIFCONF,&ifc)==-1) {
    perror("Error getting interface list");
    free(ifr);
    close(s);
    *len=0;
    return;
  }
  count=ifc.ifc_len/sizeof(struct ifreq);
  for (i=0,*len=0;i<count;i++) {
    in=(struct sockaddr_in *)&ifr[i].ifr_addr;
    strcpy(ifr_tmp.ifr_name,ifr[i].ifr_name);
    if (ioctl(s,SIOCGIFFLAGS,&ifr_tmp)==-1) {
      continue;
    }
    if (in&&(in->sin_family==AF_INET)&&((ifr_tmp.ifr_flags&IFF_UP)!=0)) {
      memset(&list[*len],0,sizeof(struct iflist));
      strncpy(list[*len].name,ifr[i].ifr_name,sizeof(list[i].name)-1);
      list[*len].addr.s_addr=in->sin_addr.s_addr;
      list[*len].isloopback=(ifr_tmp.ifr_flags&IFF_LOOPBACK)!=0;
      (*len)++;
    }
  }
  free(ifr);
  close(s);
}

#endif /*if FreeBSD*/

#ifdef VMS
pid_t GENERIC_SETSID(void) { return(0); }
#endif

#endif /*if WINDOWS*/

const char *strfunc(int func)
{
  switch (func) {
  case ANNOUNCE:
    return "ANNOUNCE";
  case REGISTER:
    return "REGISTER";
  case REG_CONF:
    return "REG_CONF";
  case FILESEG:
    return "FILESEG";
  case DONE:
    return "DONE";
  case STATUS:
    return "STATUS";
  case DONE_CONF:
    return "DONE_CONF";
  case ABORT:
    return "ABORT";
  default:
    return "UNKNOWN";
  }
}

void head_hton(struct uftp_h *header)
{
  header->uftp_id=htonl(header->uftp_id);
  header->func=htonl(header->func);
  header->tx_id=htonl(header->tx_id);
  header->seq_num=htonl(header->seq_num);
  header->section_num=htonl(header->section_num);
  header->nak_count=htonl(header->nak_count);
  header->blsize=htonl(header->blsize);
  header->pass=htonl(header->pass);
}
 
void head_ntoh(struct uftp_h *header)
{
  header->uftp_id=ntohl(header->uftp_id);
  header->func=ntohl(header->func);
  header->tx_id=ntohl(header->tx_id);
  header->seq_num=ntohl(header->seq_num);
  header->section_num=ntohl(header->section_num);
  header->nak_count=ntohl(header->nak_count);
  header->blsize=ntohl(header->blsize);
  header->pass=ntohl(header->pass);
}
 
void info_hton(struct fileinfo *info)
{
  info->fsize=htonl(info->fsize);
  info->block_total=htonl(info->block_total);
  info->section_total=htonl(info->section_total);
  info->open=htonl(info->open);
  info->rxlatency1=htons(info->rxlatency1);
  info->rxlatency2=htons(info->rxlatency2);
  info->hilargefsize=htonl(info->hilargefsize);
  info->lolargefsize=htonl(info->lolargefsize);
}
 
void info_ntoh(struct fileinfo *info)
{
  info->fsize=ntohl(info->fsize);
  info->block_total=ntohl(info->block_total);
  info->section_total=ntohl(info->section_total);
  info->open=ntohl(info->open);
  info->rxlatency1=ntohs(info->rxlatency1);
  info->rxlatency2=ntohs(info->rxlatency2);
  info->hilargefsize=ntohl(info->hilargefsize);
  info->lolargefsize=ntohl(info->lolargefsize);
}

void logfunc(FILE *logf, uint32_t id, int newline, int err, int sockerr, const char str[], ...)
{
  va_list args;
 
  if (id)
    fprintf(logf,"[%08X]: ",id);
  va_start(args,str);
  vfprintf(logf,str,args);
  va_end(args);
  if (sockerr) {
#ifdef WINDOWS
    char errbuf[300];
    FormatMessage(FORMAT_MESSAGE_FROM_SYSTEM, NULL, WSAGetLastError(),
                  0, errbuf, sizeof(errbuf), NULL);
    fprintf(logf, ": (%d) %s", WSAGetLastError(), errbuf);
    newline = 0;
#else
    fprintf(logf,": %s",strerror(err));
#endif
  } else if (err) {
    fprintf(logf,": %s",strerror(err));
  } 
  if (newline) {
    fprintf(logf,"\n");
  }
  fflush(logf);
}
