/*
 *  UFTP - UDP based FTP with multicast
 *
 *  Copyright (C) 2001-2010   Dennis A. Bush, Jr.   bush@tcnj.edu
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 3 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License along
 *  with this program; if not, write to the Free Software Foundation, Inc.,
 *  51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
 */

#include "uftp.h"

#define MAXLIST 20
#define CLIENT_ID 0x68
#define DEF_TIMEOUT 60
#define DEF_PORT 1044
#define DEF_PUB_MULTI "230.4.4.1"
#ifdef WINDOWS
#define DEF_LOGFILE "C:\\uftpd_log.txt"
#define DESTDIR "C:\\temp"
#elif defined VMS
#define DEF_LOGFILE "SYS$SCRATCH:uftpd_log.txt"
#define DESTDIR "SYS$SCRATCH"
#else
#define DEF_LOGFILE "/tmp/uftpd.log"
#define DESTDIR "/tmp"
#endif
#define DEF_LATENCY1 1500
#define DEF_LATENCY2 1500
#define MAX_INTERFACES 20
#define TEMPDIR ""
#define USAGE "uftpd [ -d ] [ -n ] [ -L logfile ] [ -T temp_dir ]\n\
    [ -D dest_dir ] [ -t timeout ] [ -p port ] [ -P pidfile ] [ -B buf_size ]\n\
    [ -M pub_mcast_addr[,pub_mcast_addr...] ] [ -I interface[,interface...] ]\n"

struct list {
  uint32_t txID;
  struct in_addr multi;
  int unicast,file,latency1,latency2;
  char filename[MAXFILENAME],*naklist;
  int32_t numblocks,numsections;
  struct sockaddr_in dest;
  f_offset_t filesize;
  int phase,attempt,minimal_resp;
  timestamp timeout_time;
  int last;
  f_offset_t prev_offset;
  int wait;
} id_list[MAXLIST];

SOCKET listener;
int timeout;
char tempdir[MAXDIRNAME],destdir[MAXDIRNAME];
int verbose,noname,interface_count;
uint16_t port;
struct in_addr m_interface[MAX_INTERFACES];
FILE *applog;

extern char *optarg;
extern int optind;

void cleanup(void)
{
  closesocket(listener);
  fclose(stderr);

#ifdef WINDOWS
  WSACleanup();
#endif
}

void gotsig(int sig)
{
  logfunc(applog,0,1,0,0,"Exiting on signal %d",sig);
  exit(0);
}

void gotpipe(int sig)
{
  logfunc(applog,0,1,0,0,"Got SIGPIPE");
}

void file_cleanup(int listindex)
{
  struct ip_mreq multi;
  int i;

  if (!id_list[listindex].unicast) {
    for (i=0;i<interface_count;i++) {
      multi.imr_multiaddr=id_list[listindex].multi;
      multi.imr_interface=m_interface[i];
      if (setsockopt(listener,IPPROTO_IP,IP_DROP_MEMBERSHIP,(char *)&multi,sizeof(multi))==SOCKET_ERROR) {
        logfunc(applog,id_list[listindex].txID,1,errno,1,"Error leaving multicast group for receiver");
      }
    }
  }
  if (id_list[listindex].file >= 0) {   /* Only close the file if not already closed */
    close(id_list[listindex].file);    /* PATCH MRL */
    id_list[listindex].file = -1;
  }
  if (id_list[listindex].naklist!=NULL) {
    free(id_list[listindex].naklist);
    id_list[listindex].naklist=NULL;
  }
  id_list[listindex].txID=0;
}

void abort_msg(SOCKET sock, struct sockaddr_in *mes_dest, int32_t id, char errstr[], int minimal_resp)
{
  char buf[PACKETSIZE],*data;
  struct uftp_h *header;
  int size;

  data=&(buf[sizeof(struct uftp_h)]);
  header=(struct uftp_h *)&(buf[0]);
  header->uftp_id=UFTP_ID;
  header->func=ABORT;
  header->tx_id=id;
  strncpy(data,errstr,BLOCKSIZE-1);
  data[BLOCKSIZE-1]='\x0';
  header->blsize=strlen(errstr);
  if (minimal_resp) {
    size=header->blsize+sizeof(struct uftp_h);
  } else {
    size=sizeof(buf);
  }
  head_hton(header);
  if (sendto(sock,buf,size,0,(struct sockaddr *)mes_dest,sizeof(*mes_dest))==SOCKET_ERROR) {
    logfunc(applog,id,1,errno,1,"Error sending ABORT");
  }
}

void set_timeout(int listindex)
{
  timestamp current_time;

  gettimestamp(current_time);
  switch (id_list[listindex].phase) {
  case 1:
    id_list[listindex].timeout_time=add_msec_interval(current_time,id_list[listindex].latency1);
    break;
  case 2:
    id_list[listindex].timeout_time=add_sec_interval(current_time,timeout);
    break;
  case 3:
    id_list[listindex].timeout_time=add_msec_interval(current_time,id_list[listindex].latency2);
    break;
  }
}

void send_register(int listindex)
{
  struct uftp_h *header;
  char buf[PACKETSIZE];
  int size;

  if (id_list[listindex].attempt==0) {
    logfunc(applog,id_list[listindex].txID,1,0,0,"Registration unconfirmed by server");
    abort_msg(listener,&(id_list[listindex].dest),id_list[listindex].txID,"Registration unconfirmed",id_list[listindex].minimal_resp);
    file_cleanup(listindex);
  } else {
    bzero(buf,sizeof(buf));
    header=(struct uftp_h *)&(buf[0]);
    header->uftp_id=UFTP_ID;
    header->func=REGISTER;
    header->tx_id=id_list[listindex].txID;
    header->blsize=0;
    if (id_list[listindex].minimal_resp) {
      size=header->blsize+sizeof(struct uftp_h);
    } else {
      size=sizeof(buf);
    }
    head_hton(header);
    if (sendto(listener,buf,size,0,(struct sockaddr *)&(id_list[listindex].dest),sizeof(id_list[listindex].dest))==SOCKET_ERROR) {
      logfunc(applog,id_list[listindex].txID,1,errno,1,"Error sending REGISTER");
    } else {
      logfunc(applog,id_list[listindex].txID,1,0,0,"REGISTER sent");
    }
    id_list[listindex].attempt--;
    set_timeout(listindex);
  }
}

void get_announce(struct sockaddr_in dest, char buf[])
{
  struct uftp_h *header;
  struct fileinfo *info;
  struct hostent *hp;
  time_t t;
  struct ip_mreq multi;
  char filepath[MAXPATHNAME];
  int listindex,i,j,found;

  header=(struct uftp_h *)&(buf[0]);
  info=(struct fileinfo *)&(buf[sizeof(struct uftp_h)]);
  info_ntoh(info);
  if (info->open==0) {
    i=0;
    found=0;
    while ((info->addr_list[i].s_addr!=0)&&(!found)) {
      for (j=0;j<interface_count;j++) {
        if (info->addr_list[i].s_addr==m_interface[j].s_addr) {
          found=1;
          break;
        }
      }
      i++;
    }
    if (!found) {
      logfunc(applog,0,1,0,0,"Name not in host list for id %08X",header->tx_id);
      return;
    }
  }
  listindex=-1;
  for (i=0;i<MAXLIST;i++) {
    if (id_list[i].txID==0) {
      id_list[i].txID=header->tx_id;
      id_list[i].multi=info->mcast;
      if (info->mcast.s_addr==0)
        id_list[i].unicast=1;
      else
        id_list[i].unicast=0;
      strncpy(id_list[i].filename,info->name,sizeof(id_list[i].filename)-1);
      id_list[i].filename[sizeof(id_list[i].filename)-1]='\x0';
      if (info->rxlatency1)
        id_list[i].latency1=info->rxlatency1;
      else
        id_list[i].latency1=DEF_LATENCY1;
      if (info->rxlatency2)
        id_list[i].latency2=info->rxlatency2;
      else
        id_list[i].latency2=DEF_LATENCY2;
      id_list[i].numblocks=info->block_total;
      id_list[i].numsections=info->section_total;
      id_list[i].dest=dest;
      if (info->fsize==0) {
        id_list[i].filesize=(f_offset_t)info->hilargefsize<<32;
        id_list[i].filesize|=info->lolargefsize;
      } else {
        id_list[i].filesize=info->fsize;
      }
      id_list[i].minimal_resp=info->minimal_resp;
      id_list[i].phase=1;
      id_list[i].attempt=5;
      listindex=i;
      break;
    }
  }
  if (listindex==-1) {
    logfunc(applog,header->tx_id,1,0,0,"Error: maximum number of incoming files exceeded: %d\n",MAXLIST);
    return;
  }
  if (!strcmp(tempdir,"")) {
    snprintf(filepath,sizeof(filepath)-1,"%s%c%s",destdir,PATH_SEP,id_list[listindex].filename);
  } else {
    snprintf(filepath,sizeof(filepath)-1,"%s%c%s",tempdir,PATH_SEP,id_list[listindex].filename);
  }
  filepath[sizeof(filepath)-1]='\x0';
  t=time(NULL);
  logfunc(applog,id_list[listindex].txID,0,0,0,"Received request at %s",ctime(&t));
  if (!noname&&(hp=gethostbyaddr((char *)&(id_list[listindex].dest.sin_addr),sizeof(struct in_addr),AF_INET))) {
    logfunc(applog,id_list[listindex].txID,1,0,0,"Request from %s (%s) id %08X",hp->h_name,inet_ntoa(id_list[listindex].dest.sin_addr),id_list[listindex].txID);
  } else {
    logfunc(applog,id_list[listindex].txID,1,0,0,"Request from %s id %08X",inet_ntoa(id_list[listindex].dest.sin_addr),id_list[listindex].txID);
  }
  logfunc(applog,id_list[listindex].txID,1,0,0,"Name of file to receive: %s",id_list[listindex].filename);
  if (sizeof(f_offset_t)>4)
    logfunc(applog,id_list[listindex].txID,1,0,0,"Bytes: %s, Blocks: %d, Sections: %d",printll(id_list[listindex].filesize),id_list[listindex].numblocks,id_list[listindex].numsections);
  else
    logfunc(applog,id_list[listindex].txID,1,0,0,"Bytes: %ld, Blocks: %d, Sections: %d",id_list[listindex].filesize,id_list[listindex].numblocks,id_list[listindex].numsections);
  logfunc(applog,id_list[listindex].txID,1,0,0,"Using private multicast address %s",inet_ntoa(id_list[listindex].multi));
  if ((id_list[listindex].file=open(filepath,OPENWRITE|O_CREAT|O_TRUNC,0644))==-1) {
    logfunc(applog,id_list[listindex].txID,1,errno,0,"Error opening data file");
    abort_msg(listener,&(id_list[listindex].dest),id_list[listindex].txID,"Error opening data file",id_list[listindex].minimal_resp);
    file_cleanup(listindex);
    return;
  }
  if (!id_list[listindex].unicast) {
    for (i=0;i<interface_count;i++) {
      multi.imr_multiaddr=id_list[listindex].multi;
      multi.imr_interface=m_interface[i];
      if (setsockopt(listener,IPPROTO_IP,IP_ADD_MEMBERSHIP,(char *)&multi,sizeof(multi))==SOCKET_ERROR) {
        logfunc(applog,id_list[listindex].txID,1,errno,1,"Error joining multicast group for receiver");
        file_cleanup(listindex);
        return;
      }
    }
  }
  send_register(listindex);
}

int addr_in_list(struct statusinfo *info)
{
  int i,j,found;

  for (i=0,found=0;(i<MAXINFODEST)&&(!found);i++) {
    for (j=0;(j<interface_count)&&(!found);j++) {
      if (info->addr_list[i].s_addr==m_interface[j].s_addr) {
        found=1;
      }
    }
  }
  return found;
}

void get_reg_conf(int listindex, char buf[])
{
  struct uftp_h *header;
  struct statusinfo *info;
  char *data;
  int i;

  header=(struct uftp_h *)&(buf[0]);
  info=(struct statusinfo *)&(buf[sizeof(struct uftp_h)]);
  data=&(buf[sizeof(struct uftp_h)]);

  if (header->func==ABORT) {
    logfunc(applog,id_list[listindex].txID,1,0,0,"Error: Transfer aborted by server: %s",data);
    file_cleanup(listindex);
  } else if (header->func!=REG_CONF) { 
    logfunc(applog,id_list[listindex].txID,1,0,0,"Invalid REG_CONF message: wrong func: %s",strfunc(header->func));
  } else if (addr_in_list(info)) {
    logfunc(applog,id_list[listindex].txID,1,0,0,"Registration confirmed");
    id_list[listindex].phase=2;
    id_list[listindex].naklist=(char *)malloc(id_list[listindex].numblocks);
    for (i=0;i<id_list[listindex].numblocks;i++)
      id_list[listindex].naklist[i]=1;
    id_list[listindex].last=-1;
    id_list[listindex].prev_offset=0;
    id_list[listindex].wait=1;
    lseek_func(id_list[listindex].file,0,SEEK_SET);
    set_timeout(listindex);
  }
}

void move_file(int listindex)
{
  char temppath[MAXPATHNAME],destpath[MAXPATHNAME];

  close(id_list[listindex].file);
  id_list[listindex].file = -1;  /* Mark the file as closed - PATCH MRL */
  if (strcmp(tempdir,"")) {
    snprintf(destpath,sizeof(destpath)-1,"%s%c%s",destdir,PATH_SEP,id_list[listindex].filename);
    destpath[sizeof(destpath)-1]='\x0';
    snprintf(temppath,sizeof(temppath)-1,"%s%c%s",tempdir,PATH_SEP,id_list[listindex].filename);
    temppath[sizeof(temppath)-1]='\x0';
    unlink(destpath);
    if (link(temppath,destpath)==-1) {
      logfunc(applog,id_list[listindex].txID,1,errno,0,"Couldn't move file");
    } else {
      unlink(temppath);
    }
  }
}

void send_done(int listindex)
{
  struct uftp_h *header;
  char buf[PACKETSIZE];
  int size;

  if (id_list[listindex].attempt==0) {
    logfunc(applog,id_list[listindex].txID,1,0,0,"Recept unconfirmed by server");
    move_file(listindex);
    file_cleanup(listindex);
  } else {
    bzero(buf,sizeof(buf));
    header=(struct uftp_h *)&(buf[0]);
    header->uftp_id=UFTP_ID;
    header->func=STATUS;
    header->tx_id=id_list[listindex].txID;
    header->nak_count=-1;
    logfunc(applog,id_list[listindex].txID,1,0,0,"Getting confirmation...");
    header->blsize=0;
    if (id_list[listindex].minimal_resp) {
      size=header->blsize+sizeof(struct uftp_h);
    } else {
      size=sizeof(buf);
    }
    head_hton(header);
    if (sendto(listener,buf,size,0,(struct sockaddr *)&(id_list[listindex].dest),sizeof(id_list[listindex].dest))==SOCKET_ERROR) {
      logfunc(applog,id_list[listindex].txID,1,errno,1,"Error sending STATUS");
    }
    id_list[listindex].attempt--;
    set_timeout(listindex);
  }
}

void send_status(int listindex, int pass, int section)
{
  struct uftp_h *header;
  char buf[PACKETSIZE],*list;
  int32_t i,l_index,n_index;
  int numnaks,done,size;
  int sectionoffset,blocksthissec;

  sectionoffset=BLOCKSIZE*8*(section-1);
  blocksthissec=((section<id_list[listindex].numsections)?(BLOCKSIZE*8):(id_list[listindex].numblocks%(BLOCKSIZE*8)));
  if (id_list[listindex].numsections && !blocksthissec) blocksthissec=BLOCKSIZE*8;
  header=(struct uftp_h *)&(buf[0]);
  list=&(buf[sizeof(struct uftp_h)]);
  numnaks=0;
  bzero(buf,sizeof(buf));
  for (i=0;i<blocksthissec;i++) {
    n_index=i+sectionoffset;
    l_index=i;
    if (id_list[listindex].naklist[n_index]) {
      if (verbose>=2)
        logfunc(applog,id_list[listindex].txID,1,0,0,"NAK for %d",n_index);
      list[l_index>>3]|=(1<<(l_index&7));
      numnaks++;
    }
  }
  if (verbose>=3) {
    for (i=0;i<BLOCKSIZE*8;i++) {
      if ((section==id_list[listindex].numsections)&&(i>=id_list[listindex].numblocks%(BLOCKSIZE*8))) {
        logfunc(applog,0,1,0,0,"");
        break;
      }
      logfunc(applog,0,0,0,0,"%d",(list[i/8] & (1<<(i%8)))!=0);
      if (i%8==7) logfunc(applog,0,0,0,0," ");
      if (i%64==63) logfunc(applog,0,1,0,0,"");
    }
  }
  done=1;
  for (i=0;i<id_list[listindex].numblocks;i++) {
    if (id_list[listindex].naklist[i]) {
      done=0;
      break;
    }
  }
  if (done&&(section>=id_list[listindex].numsections)) {
    numnaks=-1;
  } else {
    if (id_list[listindex].wait&&(numnaks==1)) {
      if (section==id_list[listindex].numsections)
        n_index=id_list[listindex].numblocks-1;
      else
        n_index=BLOCKSIZE*8*section-1;
      if (id_list[listindex].naklist[n_index]) {
        /*wait for last packet*/
        numnaks=-3;
      }
    }
  }
  id_list[listindex].wait=1;
  if (numnaks>=0) {
    header->uftp_id=UFTP_ID;
    header->func=STATUS;
    header->tx_id=id_list[listindex].txID;
    header->nak_count=numnaks;
    header->section_num=section;
    header->pass=pass;
    header->blsize=numnaks?BLOCKSIZE:0;
    if (id_list[listindex].minimal_resp) {
      size=header->blsize+sizeof(struct uftp_h);
    } else {
      size=sizeof(buf);
    }
    head_hton(header);
    if (sendto(listener,buf,size,0,(struct sockaddr *)&(id_list[listindex].dest),sizeof(id_list[listindex].dest))==SOCKET_ERROR) {
      logfunc(applog,id_list[listindex].txID,1,errno,1,"Error sending STATUS");
    } else {
      logfunc(applog,id_list[listindex].txID,1,0,0,"Sent %d NAKs for pass %d section %d",numnaks,pass,section);
    }
  } else if (numnaks==-3) {
    logfunc(applog,id_list[listindex].txID,1,0,0,"Waiting for last packet in section...");
    id_list[listindex].wait=0;
  } else if (numnaks==-1) {
    logfunc(applog,id_list[listindex].txID,1,0,0,"File transfer complete");
    id_list[listindex].phase=3;
    id_list[listindex].attempt=5;
    send_done(listindex);
  }
}

void get_fileseg(int listindex, char buf[])
{
  struct uftp_h *header;
  struct statusinfo *info;
  char *data;
  f_offset_t offset,ret_val;

  header=(struct uftp_h *)&(buf[0]);
  info=(struct statusinfo *)&(buf[sizeof(struct uftp_h)]);
  data=&(buf[sizeof(struct uftp_h)]);

  if (header->func==ABORT) {
    logfunc(applog,id_list[listindex].txID,1,0,0,"Error: Transfer aborted by server: %s",data);
    file_cleanup(listindex);
  } else if (header->func==FILESEG) { 
    if (verbose>=2) {
      if (header->pass!=1)
        logfunc(applog,id_list[listindex].txID,1,0,0,"Got packet %d",header->seq_num);
    } else if (verbose==1) {
      if (header->seq_num!=id_list[listindex].last+1)
        logfunc(applog,id_list[listindex].txID,1,0,0,"Got packet %d, last was %d",header->seq_num,id_list[listindex].last);
      id_list[listindex].last=header->seq_num;
    }
    offset=(f_offset_t)header->seq_num*BLOCKSIZE;
    if ((ret_val=lseek_func(id_list[listindex].file,offset-id_list[listindex].prev_offset,SEEK_CUR))==-1) {
      logfunc(applog,id_list[listindex].txID,1,errno,0,"lseek failed for file");
    }
    if (ret_val!=offset) {
      if (sizeof(f_offset_t)>4)
        logfunc(applog,id_list[listindex].txID,1,0,0,"offset is %s, should be %s",printll(ret_val),printll(offset));
      else
        logfunc(applog,id_list[listindex].txID,1,0,0,"offset is %ld, should be %ld",ret_val,offset);
    }
    if ((header->blsize!=BLOCKSIZE)&&(header->seq_num!=id_list[listindex].numblocks-1)) {
      logfunc(applog,id_list[listindex].txID,1,0,0,"Bad block size: %d",header->blsize);
    }
    if ((ret_val=write(id_list[listindex].file,data,header->blsize))==-1) {
      logfunc(applog,id_list[listindex].txID,1,0,0,"Write failed for segment %d",header->seq_num);
    }
    id_list[listindex].prev_offset=offset+ret_val;
    if (ret_val!=header->blsize) {
      logfunc(applog,id_list[listindex].txID,1,0,0,"Write failed for segment %d, only wrote %d bytes",header->seq_num,ret_val);
    }
    id_list[listindex].naklist[header->seq_num]=0;
    set_timeout(listindex);
  } else if (header->func==DONE) { 
    logfunc(applog,id_list[listindex].txID,1,0,0,"Got DONE message for pass %d section %d",header->pass,header->section_num);
    if (addr_in_list(info)) {
      send_status(listindex,header->pass,header->section_num);
    }
    set_timeout(listindex);
  } else if ((header->func!=REG_CONF)&&(header->func!=DONE_CONF)) {
    logfunc(applog,id_list[listindex].txID,1,0,0,"Invalid FILESEG message: wrong func: %s",strfunc(header->func));
  }
}

void get_done_conf(int listindex, char buf[])
{
  struct uftp_h *header;
  struct statusinfo *info;
  char *data;
  time_t t;

  header=(struct uftp_h *)&(buf[0]);
  info=(struct statusinfo *)&(buf[sizeof(struct uftp_h)]);
  data=&(buf[sizeof(struct uftp_h)]);

  if (header->func==ABORT) {
    logfunc(applog,id_list[listindex].txID,1,0,0,"Error: Transfer aborted by server: %s",data);
    file_cleanup(listindex);
  } else if (header->func!=DONE_CONF) { 
    logfunc(applog,id_list[listindex].txID,1,0,0,"Invalid DONE_CONF message: wrong func: %s",strfunc(header->func));
  } else if (addr_in_list(info)) {
    move_file(listindex);
    t=time(NULL);
    logfunc(applog,id_list[listindex].txID,0,0,0,"File transfer confirmed at %s",ctime(&t));
    file_cleanup(listindex);
  }
}

struct timeval *getrecenttimeout()
{
  /* go through timestamps in list
     get smallest nonzero timestamp tt
       if any are expired, do appropriate action for phase and recheck time
     if tt =< time, return 0 time
     if tt > time, return tt - time
     if all zero, return null pointer
  */
  static struct timeval tv={0,0};
  timestamp current_timestamp,min_timestamp;
  int i,found,done;
  int32_t usecs;

  gettimestamp(current_timestamp);
  done=0;
  while (!done) {
    found=0;
    done=1;
    for (i=0;i<MAXLIST;i++) {
      if (id_list[i].txID) {
        if (cmptimestamp(current_timestamp,id_list[i].timeout_time)>=0) {
          switch (id_list[i].phase) {
          case 1:
            send_register(i);
            break;
          case 2:
            logfunc(applog,id_list[i].txID,1,0,0,"Transfer timed out");
            abort_msg(listener,&(id_list[i].dest),id_list[i].txID,"Transfer timed out",id_list[i].minimal_resp);
            file_cleanup(i);
            break;
          case 3:
            send_done(i);
            break;
          }
          done=0;
          break;
        } else if (!found) {
          min_timestamp=id_list[i].timeout_time;
          found=1;
        } else if (cmptimestamp(id_list[i].timeout_time,min_timestamp)<0) {
          min_timestamp=id_list[i].timeout_time;
        }
      }
    }
  }
  if (found) {
    usecs=diff_usec(min_timestamp,current_timestamp);
    tv.tv_sec=usecs/1000000;
    tv.tv_usec=usecs%1000000;
    return &tv;
  } else {
    return NULL;
  }
}

void mainloop()
{
  struct uftp_h *header;
  char buf[PACKETSIZE];
  int rval,i,listindex;
  socklen_t addr_len;
  fd_set fdin;
  struct sockaddr_in dest;
  struct timeval *tv;

  bzero(buf,sizeof(buf));
  bzero(&dest,sizeof(dest));
  addr_len=sizeof(dest);
  header=(struct uftp_h *)&(buf[0]);
  tv=NULL;
  while (1) {
    FD_ZERO(&fdin);
    FD_SET(listener,&fdin);
    tv=getrecenttimeout();
    if ((rval=select(FD_SETSIZE-1,&fdin,NULL,NULL,tv))==SOCKET_ERROR) {
      logfunc(applog,0,1,errno,1,"Select failed");
      continue;
    }
    if (rval==0) {
      continue;
    } else if (FD_ISSET(listener,&fdin)) {
      if ((rval=recvfrom(listener,buf,sizeof(buf),0,(struct sockaddr *)&dest,&addr_len))==SOCKET_ERROR) {
        logfunc(applog,0,1,errno,1,"Error receiving in server");
        continue;
      } else if (rval!=PACKETSIZE) {
        logfunc(applog,0,1,0,0,"Invalid size %d of received packet",rval);
        continue;
      }
      head_ntoh(header);
      if (header->uftp_id==UFTP_ID) {
        listindex=-1;
        for (i=0;i<MAXLIST;i++) {
          if (header->tx_id==id_list[i].txID) {
            listindex=i;
            break;
          }
        }
        if (header->func==ANNOUNCE) {
          if (listindex==-1) {
            get_announce(dest,buf);
          }
        } else {
          if (listindex==-1) {
            /*logfunc(applog,0,1,0,0,"Invalid txID: %08X, doesn't match any current file",header->tx_id);*/
            continue;
          }
          switch (id_list[listindex].phase) {
          case 1:
            get_reg_conf(listindex,buf);
            break;
          case 2:
            get_fileseg(listindex,buf);
            break;
          case 3:
            get_done_conf(listindex,buf);
            break;
          }
        }
      } else {
        logfunc(applog,0,1,0,0,"Invalid message: not uftp packet");
      }
    }
  }
}

int main(int argc, char *argv[])
{
  struct iflist ifl[MAX_INTERFACES];
  struct ip_mreq multi;
  struct hostent *hp;
  struct in_addr *addr;
  struct sockaddr_in sin;
  int debug,c,fd,i,ifl_len,foundif;
  int32_t buffer;
  char logfile[MAXDIRNAME],pub_multi[200],*p;
  char tempf1[MAXFILENAME],tempf2[MAXFILENAME],hostname[300];
  time_t t;
  char pidfile[MAXDIRNAME];

#ifdef WINDOWS
  struct WSAData data;
  if (WSAStartup(2,&data)) {
    logfunc(stderr, 0, 1, 1, 1, "Error in WSAStartup");
    exit(1);
  }
#endif

  ifl_len=sizeof(ifl)/sizeof(struct iflist);
  getiflist(ifl,&ifl_len);
  bzero(&sin,sizeof(sin));
  bzero(&multi,sizeof(multi));
  strncpy(tempdir,TEMPDIR,sizeof(tempdir)-1);
  tempdir[sizeof(tempdir)-1]='\x0';
  strncpy(destdir,DESTDIR,sizeof(destdir)-1);
  destdir[sizeof(destdir)-1]='\x0';
  addr=NULL;
  timeout=DEF_TIMEOUT;
  port=DEF_PORT;
  debug=0;
  verbose=0;
  noname=0;
  buffer=0;
  interface_count=0;
  strcpy(pidfile,"");
  strncpy(logfile,DEF_LOGFILE,sizeof(logfile)-1);
  logfile[sizeof(logfile)-1]='\x0';
  strncpy(pub_multi,DEF_PUB_MULTI,sizeof(pub_multi)-1);
  pub_multi[sizeof(pub_multi)-1]='\x0';
  while ((c=getopt(argc,argv,"dv:nL:P:I:t:p:T:D:M:B:"))!=EOF) {
    switch (c) {
    case 'd':
      debug=1;
      break;
    case 'n':
      noname=1;
      break;
    case 'v':
      verbose=atoi(optarg);
      break;
    case 'L':
      strncpy(logfile,optarg,sizeof(logfile)-1);
      logfile[sizeof(logfile)-1]='\x0';
      break;
    case 'P':
      strncpy(pidfile,optarg,sizeof(pidfile)-1);
      pidfile[sizeof(pidfile)-1]='\x0';
      break;
    case 'I':
      p=(char *)strtok(optarg,",");
      while (p!=NULL) {
        for (i=0,foundif=0;(i<ifl_len)&&(!foundif);i++) {
          if (!strcmp(p,ifl[i].name)) {
            foundif=1;
          }
        }
        if (foundif) {
          m_interface[interface_count++].s_addr=ifl[i-1].addr.s_addr;
          p=(char *)strtok(NULL,",");
          continue;
        }
        if (inet_addr(optarg)==INADDR_NONE) {
          if ((hp=gethostbyname(p))==NULL) {
            fprintf(stderr,"Invalid host name: %s\n",p);
          } else {
            addr=(struct in_addr *)hp->h_addr_list[0];
            m_interface[interface_count++].s_addr=addr->s_addr;
          }
        } else {
          m_interface[interface_count++].s_addr=inet_addr(p);
        }
        p=(char *)strtok(NULL,",");
      }
      break;
    case 'p':
      port=atoi(optarg);
      if (port==0) {
        fprintf(stderr,"Invalid port\n");
        exit(1);
      }
      break;
    case 't':
      if ((atoi(optarg)<=0)||(atoi(optarg)>3600)) {
        fprintf(stderr,"Invalid timeout\n");
      } else {
        timeout=atoi(optarg);
      }
      break;
    case 'T':
      strncpy(tempdir,optarg,sizeof(tempdir)-1);
      tempdir[sizeof(tempdir)-1]='\x0';
      break;
    case 'D':
      strncpy(destdir,optarg,sizeof(destdir)-1);
      destdir[sizeof(destdir)-1]='\x0';
      break;
    case 'M':
      strncpy(pub_multi,optarg,sizeof(pub_multi)-1);
      pub_multi[sizeof(pub_multi)-1]='\x0';
      break;
    case 'B':
      buffer=atoi(optarg);
      if ((buffer<65536)||(buffer>104857600)) {
        fprintf(stderr,"Invalid buffer size\n");
        exit(1);
      }
      break;
    case '?':
      fprintf(stderr,USAGE);
      exit(1);
    }
  }
#ifndef VMS  /* don't use interface list on OpenVMS */
  if (interface_count==0) {
    for (i=0;i<ifl_len;i++) {
      if (!ifl[i].isloopback) {
        m_interface[interface_count++].s_addr=ifl[i].addr.s_addr;
      }
    }
  }
#endif
  /* No non-loopback interfaces, so do it the old way */
  if (interface_count==0) {
    gethostname(hostname,sizeof(hostname));
    if ((hp=gethostbyname(hostname))==NULL) {
      fprintf(stderr,"Can't get host name\n");
      exit(1);
    } else {
      addr=(struct in_addr *)hp->h_addr_list[0];
      m_interface[interface_count++].s_addr=addr->s_addr;
    }
  }
  if (!strcmp(tempdir,destdir)) {
    fprintf(stderr,"ERROR: temp dir and dest dir must be different\n");
    exit(1);
  } else if ((!isfullpath(destdir))||(!isfullpath(tempdir)&&(strcmp(tempdir,"")))) {
    fprintf(stderr,"ERROR: must specify absolute pathname for temp/dest directories\n");
    exit(1);
  } else {
    snprintf(tempf1,sizeof(tempf1)-1,"%s%cuftptmp1",tempdir,PATH_SEP);
    tempf1[sizeof(tempf1)-1]='\x0';
    snprintf(tempf2,sizeof(tempf1)-1,"%s%cuftptmp2",destdir,PATH_SEP);
    tempf2[sizeof(tempf2)-1]='\x0';
    if ((fd=open(tempf2,O_WRONLY|O_CREAT,0644))<0) {
      perror("couldn't write to dest directory");
      exit(1);
    }
    close(fd);
    if (strcmp(tempdir,"")) {
      if (link(tempf2,tempf1)==-1) {
        perror("couldn't link to from temp to dest directory");
        unlink(tempf2);
        exit(1);
      }
      unlink(tempf1);
    }
    unlink(tempf2);
  }
  atexit(cleanup);
  sin.sin_family=AF_INET;
  sin.sin_addr.s_addr=htonl(INADDR_ANY);
  sin.sin_port=htons(port);
  if ((listener=socket(AF_INET,SOCK_DGRAM,0))==INVALID_SOCKET) {
    logfunc(stderr,0,1,errno,1,"Error creating socket for listener");
    exit(1);
  }
  if (bind(listener,(struct sockaddr *)&sin,sizeof(sin))==SOCKET_ERROR) {
    logfunc(stderr,0,1,errno,1,"Error binding socket for listener");
    exit(1);
  }
  if (buffer) {
    if (setsockopt(listener,SOL_SOCKET,SO_RCVBUF,(char *)&buffer,sizeof(buffer))==SOCKET_ERROR) {
      logfunc(stderr,0,1,errno,1,"Error setting receive buffer size");
      exit(1);
    }
  } else {
    buffer=262144;
    if (setsockopt(listener,SOL_SOCKET,SO_RCVBUF,(char *)&buffer,sizeof(buffer))==SOCKET_ERROR) {
      buffer=233016;
      if (setsockopt(listener,SOL_SOCKET,SO_RCVBUF,(char *)&buffer,sizeof(buffer))==SOCKET_ERROR) {
        logfunc(stderr,0,1,errno,1,"Error setting receive buffer size");
        exit(1);
      }
    }
  }
  p=(char *)strtok(pub_multi,",");
  while (p!=NULL) {
    for (i=0;i<interface_count;i++) {
      multi.imr_multiaddr.s_addr=inet_addr(p);
      multi.imr_interface=m_interface[i];
      if (setsockopt(listener,IPPROTO_IP,IP_ADD_MEMBERSHIP,(char *)&multi,sizeof(multi))==SOCKET_ERROR) {
        logfunc(stderr,0,1,errno,1,"Error joining multicast group for listener");
        exit(1);
      }
    }
    p=(char *)strtok(NULL,",");
  }
#ifdef WINDOWS
  if (!debug) {
    if ((applog=fopen(logfile,"at"))==NULL) {
      perror("Can't open log file");
      exit(1);
    }
  } else {
    applog=stderr;
  }
#else
  if (!debug) {
    int pid;
    FILE *pidfh;

    if ((fd=open(logfile,O_WRONLY|O_APPEND|O_CREAT,0644))==-1) {
      perror("Can't open log file");
      exit(1);
    }
    if ((pid=fork())==-1) {
      perror("Couldn't fork");
      exit(1);
    } else if (pid>0) {
      exit(0);
    }
    if (strcmp(pidfile,"")) {
      /* Write out the pid file, before we redirect STDERR to the log. */
      if ((pidfh=fopen(pidfile,"w"))==NULL) {
        perror("Can't open pid file for writing");
        exit(1);
      }
      fprintf(pidfh, "%d\n", getpid());
      fclose(pidfh);
    }
    setsid();
    close(2);
    dup(fd);
    close(fd);
    for (fd=0;fd<30;fd++) {
      if ((fd!=2)&&(fd!=listener)) {
        close(fd);
      }
    }
#ifdef VMS
    chdir("SYS$LOGIN");
#else
    chdir("/");
#endif
    umask(0);
  }
  applog=stderr;
  nice(-20);
  for (c=1;c<=17;c++)
    signal(c,gotsig);
  signal(SIGPIPE,gotpipe);
  signal(SIGCHLD,SIG_IGN);
#endif
  for (c=0;c<MAXLIST;c++)
    id_list[c].txID=0;
  t=time(NULL);
  logfunc(applog,0,1,0,0,"");
  logfunc(applog,0,1,0,0,"%s",VERSION);
  logfunc(applog,0,0,0,0,"Starting daemon at %s",ctime(&t));
  do {
    mainloop();
  } while (1);
  return 0;
}
